#!"\xampp\perl\bin\perl.exe"
# server.pl (backend socketserver)
# (c) napp-it 2023,2024
use strict;
use vars qw($timeout $rsize  $comm $upsrv $max_connections %p $mem $sc $nocache $tm $line $ver $rl $cl $ml $nc $opt $haszfs $spath $cpath $on2 $r $command $allowed_cmd $allowed_ip $os $fn $on $dn $rh $hn $path $opt $filter $auth $pw $tpath $wpath $dpath @q $ip $t $in %current $client);           # public varlist


########################################################################
#################################### parameter #########################

# allowed client ip (single value or commalist)
# you can enter a complete ip like 192.168.1.2 or a substr like 192.168.1.
# if you enter a substr, all clients are allowed whose ip begins with this value
$allowed_ip="";

# Client timeout (read data from client, default 20)
$timeout=90;

# max client connections (read data from client, default 20)
$max_connections=25;

# callback method enables encrypted https transfers for commands and results (slower) instead unencrypted (faster)
#upsrv (Upload https Server for callback method)

# $upsrv= ";             # no callback or initiated by frontend web-gui
# $upsrv="ip";           # callback is enforced with calling ip (frontend Apache) as unvalidated https server (self signed certificate)
# $upsrv="hostname.dns"; # ex www2.universty.org, enforces a validated https server (public certificate)
$upsrv="";


##################################/ parameter ###########################
#########################################################################

################# do not edit below #######

# scriptname: server.pl

# this socket_serverscript server.pl should run to BSD, Illumos/Solaris, Linux, OSX and Windows
# for remote management from a Windows client with the cs web-gui

# start the ZFS backend server part:
# copy the folder ./[web-gui]/data/remote_cs_server to any location on a supported platform ex to /tank/data
# and start cs-backend server (as admin/root) via perl /path_to/remote_cs_server/start_server_as_admin.pl

# Add this server to a socketgroup in the frontend client Windows ZFS gui in menu Extensions > Socketgroup
# the auth key here in /socket/server.auth must be the same as on Windows in /path/_log/groups/hostname~ip.txt


###############################
$ver="Socketserver v.24.05.20";
###############################

#lib and path are set by startscript
use lib "/var/web-gui/cs_server/CGI";

# path to serverscript
$spath="/var/web-gui/cs_server";

#tmp folder cache files on remote server
$cpath="$spath/tmp";                         # cachepath

# localhost tmp => /xampp/web-gui/tmp  (localhost cache files)
if ($spath=~/\/xampp/i) {
   $cpath="/xampp/web-gui/tmp";              # cachepath
}

#remote.log for background tasks
$rl="$spath/tmp/remote.log";

#command.log
$cl="$spath/tmp/command.log";

#monitor.log
$ml="$spath/tmp/monitor.log";

if (!($max_connections > 0)) { $max_connections = 20; }      # defaults
if (!($timeout > 0))         { $timeout = 90; }

print "\ntimeout:$timeout\nmax_connections:$max_connections\nallowed_cmd:$allowed_cmd\nallowed_ip:$allowed_ip";

use IO::Socket::INET;                                        # socket communication
use IO::Select;

use HTTP::Tiny;                                              # send large data >32K via post
use URI::Escape;                                             # url escape ex " " -> %20

# other modules
#use File::Stat;                                              # file stat like age
use File::Find;                                              # find files or directories

use Fcntl qw(:flock);                                        # filelocking
use IPC::Run qw( run timeout );                              # system() alternative with stdout, stderr and timeout

use Digest::SHA::PurePerl qw(sha256_hex);                    # replaces md5m, $hash = sha256_hex($data); 96 char

use utf8;

# auto-flush on socket
$| = 1;

# Configurable parameters
#################################################

# set in config.pl
#my $allowed_cmd="";                                          # which console apps are allowed beside zfs and zpool
#my $allowed_ip="";                                           # which clients are allowed (commalist possible), ex 192.168.1.3,172.16.1.
                                                              # client ip must begin with these values or is denied
#my $max_connections="20";
#my $timeout="30"

my $port = 63000;
my $chunk_size = 4*1024;                                     # only when using chunks
                                                              # a socket transfer is limited to 32K to avoid buffer problems
# main #
############# some infos #########################
print "\n" x 15;                   # some newlines
print "$ver\n\n";


#################################
print "#######################\n";
print "initiate $ver with max $chunk_size transfers in a cmd shell \n";
print "request commands in client ex via 'zfs .. auth:=xx' or 'getdisks auth:=xx'\n\n ";
#################################


# read ps and base infos #########################
##################################################

 &get_pw;         # print "pw=$pw\n"; # debug
 &my_os;          # detect os,on,hn,rh, path and print OS info

&socket_inet();   # start socket server
exit;

#####################################################################
sub socket_inet {      #ok, socket via Socket::INET (Socket::IP has problems)
################
   #
   print "use IO::Socket::INET\n";

   # intiate socket server
   my $server = IO::Socket::INET->new(
       LocalPort => $port,
       Proto     => 'tcp',
       Listen    => $max_connections,
       Blocking  => 0,  # Set to non-blocking mode
   ) or die "Cannot create socket: $!";

  my $select = IO::Select->new($server);

  while (1) {

       foreach my $client ($select->can_read($timeout)) {
           if ($client == $server) {
               my $new_client = $server->accept;
               $select->add($new_client);

               my $ip=$new_client->peerhost;

           } else {

               my $bytes_read = $client->sysread(my $message, $chunk_size);
               if ($bytes_read) {

                   my $ip=$client->peerhost;
                   if (&client_ip_allowed($ip) eq "0")  {

                          print __LINE__." client $ip not allowed\n";
                          $client->syswrite("your ip is not allowed\n")

                   } else {

                      #  client is allowed
                      my ($reply,$filename) = &process_command($message, $ip);

                      $t=substr("$reply,$filename",0,80)." ..";
                      print "\n".__LINE__." ###########\nsocketserver_in:\n###############\n$message\n\n####reply $rsize #####\n$t\n###############\n";

                      if ($filename=~/\w/) {           # 192.._abc123.data or upload error:

                          print __LINE__." uploaded $rsize data to file $filename\n";
                          $client->syswrite("$filename\n");    # return filename or upload error

                      } else {

                          print __LINE__." return  $rsize data or error via socket\n";
                          $client->syswrite("$reply\n");       # return data
                      }
                    }

               } else {
                   print "\n#####################\n".__LINE__." Connection closed by " . $client->peerhost . "\n";
                   print "$ver\n\n";
                   $select->remove($client);
                   $client->close;
               }
           }
       }
   }

   # Beendigung der Schleife
   $server->close;

}  #/inet


######################
sub process_command {
######################
  # a request is either from 127.0.0.1 (Windows frontend to itself, unresticted) or

  # remote server call from frontend with ip
  # stat commands are processed directly. All other commands and results can be "direct" or "callback" type vih https (upsrv defined)
  # cmd for callback  rcmd::upsrv::filename::memberip::auth:::opt(nocachefrontendtime)::

  my ($message, $ip) = @_;
  my $q=$message;                                               # ip is calling host
  my (@p,$up,$auth,$cfn,$mem,$ftime,$r);

  #get param -> %p  for alias commands

  @p=&parameter($message,$ip);                                  # cfn,upsrv,auth,mem,nocacheftime
  $cfn=$p[0];
  $up=$p[1];                                                    # uploadserver, calling ip or remote hostname from settings
  $auth=$p[2];
  $mem=$p[3];                                                   # this ip from frontend view
  $nocache=$p[4];
  $ftime=$p[5];
  $q=~s/ *line:=.*//s;                                          # remove cmd options (lime is first option

  # auth and allowed ip check, not localhost
  if ($ip ne "127.0.0.1")   {                                   # local check or always allowed?
    if ($p[2] ne $pw)    {                                      # p(2)=auth
       print __LINE__." pw: $pw ne auth\n $p[2]\n";             # debug
       return('error: auth:=socket auth error');
    }

    # client_ip_allowed is already checked
  }
  ###

  if ($q eq "") {                                              # error
       print __LINE__." <- [no cmd]\n";
       return ("error: [no cmd]");
  }
                                    # insert auth to curl cmd
  #print __LINE__." process command> $q\n------\n";              # see print ipc>

  # server runs in a cmd shell
  # commands on Windows are mostly using powershell
  # to execute powershell commands ex get-disk
  # you must execute: powershell -command get-disk
  # ex @q = qw(powershell -command Get-PhysicalDisk | Select DeviceID,UniqueId);

  # always use powershell, otherwise replace

  # short status commands always allowed after auth and ip check
  if (!($q=~/^hostname|^hn|^rh|^on|^hi|^os|^dn|^ip|^fn|^alive|^mem|^sc/)) {

    # too old, del
    $t=&my_age();      # age of activity.log max 7 days
    if ($t > 3600*24*7) {
      unlink ($cl);
      unlink ($rl);
    }

    $t=&get_mydate;
    &my_add($cl,"$t \t$q");                # add q to command.log
  }

  # these short cmds always processed directly after auth verification
  if ($q eq "hostname") { print __LINE__."\n ok: <- $hn\n";       return ($hn);         }            # hostname
  if ($q eq "hn")       { print __LINE__."\n ok: <- $hn\n";       return ($hn);         }            # hostname
  if ($q eq "mem")      { print __LINE__."\n ok: <- $mem\n";      return ($mem);        }            # member ip
  if ($q eq "rh")       { print __LINE__."\n ok: <- $rh\n";       return ($rh);         }            # remote host
  if ($q eq "hi")       { print __LINE__."\n ok: <- $rh,$os\n";   return ("$rh,$os");   }            # remote host info =rh + os
  if ($q eq "os")       { print __LINE__."\n ok: <- $os\n";       return ($os);         }            # os
  if ($q eq "sc")       { print __LINE__."\n ok: <- $sc\n";       return ($sc);         }            # samba config
  if ($q eq "on")       { print __LINE__."\n ok: <- $fn\n";       return ($fn);         }            # on same as fn
  if ($q eq "dn")       { print __LINE__."\n ok: <- $on\n";       return ($on);         }            # distri name
  if ($q eq "fn")       { print __LINE__."\n ok: <- $on\n";       return ($on);         }            # family name
  if ($q eq "alive")    { print __LINE__."\n ok: <- yes,alive\n"; return ("yes,alive"); }            # alive?
  if ($q eq "ip")       { print __LINE__."\n ok: <- $ip\n";       return ($ip);         }            # ip of client


  #################  end of direct actions beside localhost, others may need callback ##


  if ($ip ne "127.0.0.1") {             # can be a direct call or download callback request
  #########################

      # if $up eq "", no callback $cmd  or up is enforced locally or can be requested from frontend  -> callback (download cmd)
      # result is always uploaded

      # download cmd?
      if (($up ne "") || ($q=~/^rcmd::/))  {           # callback, q=rcmd::filename::auth::upsrv::
                                                       # cmd for callback  rcmd::upsrv::cfn::memberip::auth:::opt(nocachefrontendtime)::
          $q=&my_download("$up","$cfn.rcmd","$auth",$ip);
          if ($q=~/error:/s) { return ($q,""); }       # reply,filename
      }

      @p=&parameter($message,$ip);                                  # cfn,upsrv,auth,mem,nocache/ftime
      $cfn=$p[0];
      $up=$p[1];                                                    # uploadserver, calling ip or remote hostname from settings
      $auth=$p[2];
      $mem=$p[3];                                                   # this ip from frontend view
      $nocache=$p[4];
      $ftime=$p[5];
      $q=~s/ *line:=.*//s;                                          # remove cmd options (lime is first option

      $t=substr($q,0,70); $t=~s/\n/<n>/gs;
      print __LINE__." remote server> $t ..\n";

      # now we have a valid command either direct or via callback

  }

  ###########################################################
  ## this is localhost execution or after .rcmd download ##
  ############################################################

  my $ps="1";                        # use powershell
  my $grep="";                       # simualate a basic grep filter with grep a or grep a,b (a and b)

  if ($q=~/\| *grep/) {              # ex tasklist | grep perl   or grep perl,nc   or grep 1234,1235
       $grep="$q";
       $grep=~s/.*grep +//;          # get filter
       $q=~s/ *\| *grep.*//;         # remove grep
       print __LINE__." grep filter: $grep\n";
  }

  # check auth
  if ($p[2] ne $pw)    {
       print __LINE__." pw: $pw ne auth $p[2]\n";       # debug
       return('error: auth error');
  }
  ################# process cmd ####################

  # background task ex zfs send snap | nc ip port &
  if ($q =~/\&$/) {

     $r=&background($q);
     return ($r);                     # diff pid

  } else {

     if ($on=~/mswin/i)  {              # windows, not darWIN

       if ($q=~/^type /)  { $q=~s/^type /get-content /;    }    # replace with powershell command
       if ($q=~/^ps axw/) { $q=~s/^ps axw */tasklist/; $ps=0; } # replace with cmd command
       if ($q=~/^curl/)   { $ps=0; }                             # curl in powershell ?

       @q=();

       if  ($ps) {                                                  # use powershell
           #
           if ($q=~/^curl/) {
                  $t=$q;
                  $t=~s/^curl *//;
                  # curl --connect -- connect.pl
                  # Start-Process -FilePath 'curl' -ArgumentList 'curl --connect-timeout 3 -X POST -k -F "upfile=123.rcmd" -F "auth=345" -F "data=YOUR_DATA_PARAMETER" https://localhost/cgi-bin/cs/cs_connect.pl' -NoNewWindow
                  $q[0]="powershell";
                  $q[1]="Start-Process";
                  $q[2]="$t";              # curl parameter
                  $q[3]="-NoNewWindow";

           } else {
                #
                $q=~s/^powershell +|^ps +//;                         # no double powershell  (from client)
                $q=~s/^ *\-command +//;                              # no double -command
                $q[0]="powershell";                                  # use array for irc::run
                $q[1]="-command";

                $q[2]="$q";                                          # remove opt
            }

          } else {

             @q=();
             @q=split(/ /,$q);                                     # use cmd
          }

     } else {  #/not win

       @q=();                                                    # crash in omnios with replacements on modify call
       @q=split(/ /,$q);
     }

  }  # /background task                                          # powershell command
  #

  if ($q eq "stat")     { print __LINE__."\n ok: <- todo\n";      return ('todo');      }            # basic stat
  my $out1="";

  if ($q eq "rl")       {
      print __LINE__."\n ok: rl <- remote.log\n";
      $out1=&my_read("$rl");
      if (!($out1 =~/\w/s)) {
        $out1='empty';
      }
  }

  if ($q eq "cl")       {
      print __LINE__."\n ok: cl <- command.log\n";
      $out1=&my_read("$cl");
      if (!($out1 =~/\w/s)) {
        $out1='empty';
      }
  }                # command log
  if ($q eq "rld")      { print __LINE__."\n ok: rld <- remote.log del\n";  $out1=&my_read("$rl");  unlink($rl);  }   # remote log del
  if ($q eq "cld")      { print __LINE__."\n ok: cld <- command.log del\n"; $out1=&my_read("$cl");  unlink($cl);  }   # command log del
  if ($q eq "ml")       { print __LINE__."\n ok: ml <- monitor.log\n";      $out1=&my_read("$ml");   }                # monitor log
  if ($q eq "mld")      { print __LINE__."\n ok: mld <- monitor.log del\n"; $out1=&my_read("$ml");  unlink($ml);  }   # monitor log del


  #append or write
  if ($q=~/^writefile/)  { $t=&write_file($q);  print "write:$t\n";  return($t); }           # writefile  ex /etc/samba/smb.conf
  if ($q=~/^appendfile/) { $t=&append_file($q); print "append:$t\n"; return($t); }           # appendfile ex /etc/pam.conf

  if (($q=~/zfs |zpool /) && ($haszfs eq "no"))  {
         print __LINE__."\n ok: <- zfs not installed\n";
         return ("zfs not installed");            # zfs not installed
  }

  #
  # filename for upload data

  my $filename="$p{'cfn'}";          # 127.0.0.1_zpool_list..
  my $lfilename=$filename;

  $lfilename.=".last";
  $filename.=".data";                #  local cachefile

  # del cache
  if ($q eq "dc")       {
     my @f = glob("$cpath/*.*");     # cachepath
     foreach my $f (@f) {
        if (!($f=~/\.data$|\.last$/)) {  next; }
        unlink($f);
     }
     print __LINE__."\n ok: <- dc\n";        return ('ok, cache files deleted');
  }


  if ((-f "$cpath/$filename") && ($nocache ne "1")) {           # use cache file

       #check last (time()\fullcmd)

       my @last=&my_read($lfilename,'a');      # time\ncmd
       $tm=time();

       if ($last[0]-$tm < 300) {               # cache 5min
             print __LINE__." could use cachefile, todo: $filename\n";
             #$out1=&my_read($filename);
       } else {
             unlink($filename);
             unlink($lfilename);
       }
  }

  if ($nocache) { print __LINE__." cache disabled\n"; }         # this is a command option: cmd opt:=nochache auth:=123

  # run cmd -> file as size of $out is limited, return only error on small io directly
    my $err="";
    my $out="";
    my $in="";

  $t=join(" ",@q);
  print __LINE__." ipc cmd> $t $in\n-------------\n";             # opt, socket in is usually enough

  if ($out1 eq "") {           # no log or cache
    ###############################################
    eval {
      if ($q[0] eq "curl") {                 # problems with ipc<..run ?
         $out=`$t`
      } else {
        run \@q, \$in, \$out, \$err, timeout( $timeout );          # execute via ipc::run and timeout (in s)
      }
    };
    ################################################
  }   else {
     $out=$out1;  # log or cache files
  }
  #;

  if ($q=~/^wmic /) {                # win replacement of ps axw, gives special char
      $out=~s/[^a-zA-Z0-9\n\t\r \.,;\-\_\#]//gs;
  }

  if ($@) {
      # eval error
      $err=$@;
      print __LINE__." eval error: $@\n";
  }


  $out=~s/\n+$//s;             # no trailing nl
  $err=~s/\n+$//s;             # no trailing nl

  if ($err =~/error:/) {

           print __LINE__."\n <- error: \n$err\n";
           return ("error: $err");  # stderr

  } else {

           # simulate grep
           if ($grep ne "") {
                my $ok;
                my @greps=split(/,/,$grep);                           # allow commalist based on and
                print __LINE__." grep $grep\n";

                my @out=split(/\n/,$out);
                my @filtered=();
                foreach my $o  (@out) {

                  $ok="";
                  foreach my $g (@greps) {                            # grep as commalist a,b -> a and b
                    #print "grep $o -> $g\n";
                    if (!($o=~/$g/)) { $ok.="0";  }                    # not found
                  }

                  if (!($ok=~/0/))    {
                     push(@filtered,$o);                               #every grep in outpur line
                  }
                }
                $out=join("\n",@filtered);
           }

           #utf-8 ?
           #utf8::encode($out);              # umlaut wrong at console

           #if ($t=~/^(?:[\x00-\x7F]|\xC0[\x80-\xBF]|\xE0[\x80-\xBF][\x80-\xBF]|\xF0[\x80-\xBF][\x80-\xBF][\x80-\xBF])*$/) {
           #  print __LINE__." likely utf-8\n";
           #} else {
           #  print __LINE__." propably not utf-8\n";
           #}

           # save result as file (caching and file upload
           $rsize=int(length($out)/1024);                        # result size
           if ($rsize < 1) { $rsize=1; }
           $rsize.="K";

           # write filename.data for caching
           &my_write("$cpath/$filename",$out);                    # cfn.data

           # write lfilename.last for caching
           $tm=time();
           &my_write("$cpath/$lfilename","$tm\n$q\n");            # cfn.last

           # upload, never on localhost (web-gui can read tmp folder)
           # for a remote server upload when $upsrv is defined or whwn size > 8K

           if ($ip ne "127.0.0.1")  {
              print __LINE__." up=$up\n";

              if (($up ne "") || (length($out) >= 1024*4)) {          # 4k, truncated transfer if larger ??
                   #
                   $r="";
                   if ($up eq "") { $up=$ip; } # fallback option
                   eval {
                      $r=&my_upload($up,$filename,$auth,$ip);           # upload file via  curl: up, filename, auth, ip
                   };

                   if ($@) {
                           #   eval error
                           $err=$@;
                           print __LINE__." eval error: $@\n";
                   }

                   ## return filename  or error from my_upload:
                   if ($r=~/error:/s)  {
                          return($r,"");
                   } else {
                          return("",$filename);
                   }

              } else {      # return data directly with send marker to detect partly transmissions

                    $t=substr($out,0,600);  $t=~s/\n| +/ /gs;
                    print __LINE__." return data $rsize: $t ...\n";
                    return("<BoS:>$out<EoS:>");                # with BoS/EoS Beginofsend, endofsend marker to detect incomplete transfers
              }

           } else {

              #localhost, return data or filename

              if (length($out) >= 1024*4) {
                 return("",$filename);
              } else {
                 return("$out",'');
              }
           }
  }
}

#############
sub parameter {  # command with parameter line:= opt:= upsrv:= rauth:= cfn:= auth:=
#############

   my $cmd=$_[0];
   my $ip=$_[1];

   $cmd=~s/\n|\r//gs;

   my ($t,$p);

   # cmd for callback  rcmd::upsrv::cfn::memberip::auth:::opt(nocachefrontendtime)::
   #                      0    1      2         3     4     5
   if ($cmd=~/^rcmd::/) {

       my @par=split(/::/,$cmd);
       $p{'upsrv'}=$par[1];
       $p{'cfn'}=$par[2];                   # command filename, no ext
       $p{'cfn'}=~s/\.data$|\.rcmd$//;
       $p{'mem'}=$par[3];
       $p{'auth'}=$par[4];
       $p{'opt'}=$par[5];                   # nocache12344
       $p{'nocache'}=$par[5];
       $p{'ftime'}=$par[5];                 # frontend time 12345

   } else {        # regular cmd like zpool list line:=...

       $t=$cmd;     if ($t=~/line:=/)     { $p{'line'}=$t;    $p{'line'}=~s/.+ *line:=//;       $p{'line'}=~s/ .+//;  }
       $t=$cmd;     if ($t=~/opt:=/)      { $p{'opt'}=$t;     $p{'opt'}=~s/.+ *opt:=//;         $p{'opt'}=~s/ .+//;  }
       $t=$cmd;     if ($t=~/mem:=/)      { $p{'mem'}=$t;     $p{'mem'}=~s/.+ *mem:=//;         $p{'mem'}=~s/ .+//;  }      # member ip from frontend view
       $t=$cmd;     if ($t=~/upsrv:=/)    { $p{'upsrv'}=$t;   $p{'upsrv'}=~s/.+ *upsrv:=//;     $p{'upsrv'}=~s/ .+//;  }
       $t=$cmd;     if ($t=~/auth:=/)     { $p{'auth'}=$t;    $p{'auth'}=~s/.+ *auth:=//;       $p{'auth'}=~s/\s.*//s;  }

       $p{'cfn'}=&c_filename($cmd,$p{'mem'});        # cfn from command without .ext
   }

   if ($p{'opt'} =~/nocache/)  {

           #del cache
           my @files = glob("$cpath/*.*");
           foreach my $f (@files) {
              if ($f=~/\.log/) { next; }
              unlink ($f);
           }

   }

   # callback
   if ($p{'upsrv'} eq "ip") {
       $p{'upsrv'}=$ip;                    # remote cmd can be "", 'ip' or 'hostname.dns'
   }

   #local is mandatory
   if ($upsrv=~/[a-z]/)  {
       if ($upsrv eq "ip")    {            # upserver is localhost Apache or  host given by About > Setting
           $p{'upsrv'}=$ip;
       } else {
          $p{'upsrv'}=$upsrv;              # enforced by app setttings, either
       }
   }   # otherwise use cmd settings

   $p{'nocache'}=~s/\d//g;                 # nocache
   $p{'ftime'}=~s/\D//g;                   # frontend time 12345

   $p{'auth'}=~s/[^a-z0-9]//g;
   print __LINE__." sub parameter: cmd=$cmd\npar: cfn=$p{'cfn'},\nupsrv=$p{'upsrv'}, auth=$p{'auth'}, mem=$p{'mem'}, nocache=$nocache\n\n";

   return($p{'cfn'},$p{'upsrv'},$p{'auth'},$p{'mem'},$p{'nocache'},$p{'ftime'});   # cfn,upsrv,auth,memberip,nocache,frontendtime
}

##############
sub append_file {
##############
# remote request to append data to file ex /etc/pamconf
# command: appendfile::file::data::EndOfFile::

my $data=$_[0];
my @parts=split(/::/,$data);
if ($parts[0] ne "appendfile")            {  return ("error: no appendfile command"); }
if (!($parts[1]=~/\/\w|\\\w/))            {  return ("error: no path+filename"); }
#if (!($parts[1]=~/.conf/)) {  return ("error: file not allowed"); }   # smb.conf or smb4.conf
if ($parts[2] eq "")                      {  return ("error: nodata"); }
if ($parts[3] ne "EndOfFile")             {  return ("error: EndOfFile not found"); }

#keep default
if (!(-f "$parts[1].default")) {
  `cp $parts[1] $parts[1].default`;           # ex pam.conf.default
}

# single line transfer
$parts[2]=~s/<n>/\n/gs;                       # single line transfer

# edit in /tmp
print __LINE__." cp $parts[1] /tmp/append.tmp\n";
`cp $parts[1] /tmp/append.tmp`;

print __LINE__." appendfile $parts[2]\n";

&my_append("/tmp/append.tmp","$parts[2]");    # add tmp file
`cp /tmp/append.tmp $parts[1]`;               # copy back
`rm /tmp/append.tmp`;                         # rm tmp
return ("ok");
}

##############
sub write_file {
##############
# remote request to write file ex /etc/samba/smb.conf
# command: writefile::file::data::EndOfFile::

my $data=$_[0];
my @parts=split(/::/,$data);
if ($parts[0] ne "writefile")             {  return ("error: no writefile command"); }
if (!($parts[1]=~/\/\w|\\\w/))            {  return ("error: no path+filename"); }
if (!($parts[1]=~/smb\.conf|smb4\.conf/)) {  return ("error: file not allowed"); }   # smb.conf or smb4.conf
if ($parts[2] eq "")                      {  return ("error: nodata"); }
if ($parts[3] ne "EndOfFile")             {  return ("error: EndOfFile not found"); }

#keep default
if (!(-f "$parts[1].default")) {
  `cp $parts[1] $parts[1].default`;        # smb.conf.default
}

# single line transfer
$parts[2]=~s/<n>/\n/gs;                    # single line transfer

print __LINE__." writefile parts[1]\nparts[2]\nparts[3]\n";
&my_write("$parts[1]","$parts[2]");    # write file
return ("ok");

}

##################
sub get_smb_conf {
##################
   my ($r,$t);

   # search smb.conf or smb4.conf
   $r=`find /etc/samba -name smb.conf`;                            # default Linux
   chomp($r);
   if ($r=~/smb.conf/s) {
       $t=&my_read($r);
       if ($t=~/workgroup/s)  {   # valid config
            return($r);
       }
   }

   $r=`find /usr -name smb.conf`;                                  # TrueNas Linux, /usr/share/samba/smb.conf
   chomp($r);
   if ($r=~/smb.conf/s) {
       $t=&my_read($r);
       if ($t=~/workgroup/s)  {   # valid config
            return($r);
       }
   }

   $r=`find /usr/local -name smb4.conf`;                            # Free-BSD 14 /usr/local/etc/
   chomp($r);
   if ($r=~/smb4.conf/s) {
       $t=&my_read($r);
       if ($t=~/workgroup/s)  {   # valid config
            return($r);
       }
   }

   $r=`find /opt -name smb.conf`;                                  # omnios
   chomp($r);
   if ($r=~/smb.conf/s) {
       $t=&my_read($r);
       if ($t=~/workgroup/s)  {   # valid config
            return($r);
       }
   }
}

############
sub get_pw {
########### get pw hash

  my @data = ();
  $wpath=$spath;
  $wpath=~s/cs_server/cfg/;
  my $file="$wpath/server.auth";

  if ($^O=~/Win32/i) {
     $t=$spath; $t=~s/data.*/cfg/;     #/xampp/web-gui/cfg
     $file="$t/server.auth";
  }
  my (@p,$data);

  if (!(-f "$file")) { #same dir as server.pl
     print __LINE__." required $file not found\n";
     exit;
  }

  print __LINE__." read auth: $file\n";
  $data=&my_read($file);

  # any string

  $data=~s/.*value_adminpw/value_adminpw/s;      # old, now use simple string in server.auth
  $data=~s/\n.*//s; $data=~s/\s//gs;             # first line, no spaces

 # print "$data";
  $pw=$data;

  return(1);

}

################    command1 & or command1 | command2 &
sub background {
################

   my $command=$_[0];

   if (!($command =~/\&$/)) { print __LINE__." sub background: $command not a background job\n"; return; }

   $command=~s/ +$//;
   $command=~ s/\P{ASCII}//g;                    # remove non accii 127  char

   # ex for zfs send snap | nc ip 7000
   # Win: such a pipe does not work in powershell for unknown reason: start via cmd and start /B command

   ### file errorlog 2>&1 ?
   if ($on=~/mswin/i)  {                    # windows, not darWIN   / -> \
      $rl=~s/\//\\/g;
   }
   $command=~s/\&/ >> $rl 2>\&1 \&/;         # & -> >> remote.log 2>&1  &
   $command=~s/\bnc /$nc /;                 # must use nc from tools

   $t=&get_mydate;
   &my_add("$rl","$t");                     # date and time
   &my_add("$rl"," \t$command\n");          # time command \n
   &my_add("$rl","re: \t$line \t");         # re 123  (return value of replications)

   my $cmd1=$command;
   my $cmd2="";
   my ($r,@p,$pids1,$pids2,@v,$id);

   if ($cmd1=~/\|/) {
     $cmd1=~s/ *\|.*//;
     $cmd1=~s/ .*//; # ex zfs send -> zfs

     $cmd2=$command;                     # after pipe
     $cmd2=~s/.*\| *//;
   }

   print __LINE__." start background task $command with cmds \n$cmd1 $cmd2\n";

   #get prozesses for cmd1 and cmd2 prior start
   if ($on=~/mswin/i)  {              # windows, not darWIN


      # Abbildname                     PID Sitzungsname       Sitz.-Nr. Speichernutzung
      # ========================= ======== ================ =========== ===============
      # System Idle Process              0 Services                   0             8 K
      # ..
      # services.exe                  1352 Services                   0        10.984 K
      eval {
        $r=`tasklist`;
      };
      @p=split(/\n/,$r);
      foreach my $p (@p) {
         @v=split(/ +/,$p);
         if ($v[0]=~/zfs|nc/) { $pids1.="$v[1],"; }            # check only nc and zfs
      }

   } else {

      #
      # ps axw
      # PID TT  STAT       TIME COMMAND
      # 557  -  Ss      0:00.95 /sbin/devd
      # 758  -  Ss      0:02.35 /usr/sbin/syslogd -s
       eval {
         $r=`ps axw | grep -v ps`;
       };
       @p=split(/\n/,$r);
       foreach my $p (@p) {
         $p=~s/^ +//;
         @v=split(/\s+/,$p);
         if ($v[4]=~/zfs|nc/) {         # check only zfs and nc
           $pids1.="$v[0],";
         }
       }
   }
   $pids1=~s/,+$//;                  # processes prior start of background job
   #print __LINE__." processes prior start: $pids1\n";

   eval {

      if ($on=~/mswin/i)  {            # windows, not darWIN
          $command=~s/\&$//;            # remove &
          print "\n".__LINE__." start background job: \n########################################################\n\n";
          #system("start /B $command");  # start in background via cmd -> always in background in new console

          #powershell -command "start-process C:\xampp\perl\bin\perl.exe cs_server\server.pl -WindowStyle minimized"

          print "powershell -command \"start-process \"cmd /c start $command\"\" -WindowStyle minimized\"\n";
          system("powershell -command \"start-process \"cmd /c start $command\"\" -WindowStyle minimized\"");

      } else {

        print "\n".__LINE__." start background job: \n############################\n\n$command\n\n############################\n\n";
        system("$command");          # has & at the end
      }
   };

   sleep(1); # wait for start
   print "ok, go on\n";

   #get prozesses for cmd1 and cmd2 after start
   if ($on=~/mswin/i)  {              # windows, not darWIN

      # Abbildname                     PID Sitzungsname       Sitz.-Nr. Speichernutzung
      # ========================= ======== ================ =========== ===============
      # System Idle Process              0 Services                   0             8 K
      # ..
      # services.exe                  1352 Services                   0        10.984 K
      eval {
        $r=`tasklist`;
      };
      @p=split(/\n/,$r);
      foreach my $p (@p) {
         @v=split(/ +/,$p);
         if ($v[0]=~/zfs|\bnc/) { $pids2.="$p,"; }
      }

   } else {
      #
      # ps axw
      # PID TT  STAT       TIME   COMMAND
      # 557  -  Ss      0:00.95   /sbin/devd
      # 758  -  Ss      0:02.35   /usr/sbin/syslogd -s

       eval {
         $r=`ps axw | grep -v ps`;
       };
       @p=split(/\n/,$r);
       foreach my $p (@p) {
         $p=~s/^ +//;
         @v=split(/ +/,$p);
         if ($v[4]=~/zfs|\bnc/) {         # check only zfs and nc
           $pids2.="$p,";
         }
       }
   }

   $pids2=~s/,+$//;                  # processes prior start of background job
  # print __LINE__."\nprocesses after start: $pids2\n";

   # get difference
   my $pid_new="";
   my @p2=split(/,/,$pids2);
   foreach my $v (@p2) {
       if ($pids1=~/\b$v\b/) {  next; } # pid was there prior start, we need only new ones
       $pid_new.="$v,";
   }

   $pid_new=~s/,+$//;
   print __LINE__." new processes after start: \n$pid_new\n\n";
   &my_add($rl,"new processes after start: \n$pid_new\n");
   return ($pid_new);
}



sub my_os {
##########
  # public os, hn, rh, path
  my ($t,@t,$uname,$rel);
  print "\n";

  $haszfs="no";
  $nc="nc";                                    # netcat

  # host and os name
  $hn=`hostname`; chomp($hn);                  # X and Windows
  $dn = $^O;                                   # distribution name from perl
  $on = $dn;                                   # os familie name ex all bsd=*bsd, all Linux lx

  #defaults
  $os  ="other";
  $rh  ="OTH:$hn";
$tpath="/xampp/web-gui/tmp";


  if ($on=~/Win32/i) {

      $t=`zpool list 2>&1`;
      if (!($t=~/\"zpool\"/s)) { $haszfs="yes"; } # missing "zpool" or zfs installed

      my $ver=`ver`;                        # Microsoft Windows [Version 10.0.19045.4046]

      $t=$ver;
      $t=~s/.*Version +//is;                # 10.0.1..

      $rh="WIN:$hn";                        # W1N:hostname

      $os=$ver;
      $nc="$spath/tools/nc/win/nc64";
      $nc=~s/\//\\/g;                      # windows cmd

  } else {

      # set nc to 755
      `chmod -R 755 $spath/*`;            # mainly nc

      $uname=`uname -mrs`;                # os, kernel, cpu
      chomp($uname);

      @t=split(/ +/,$uname);              # split

      chomp($path);

      #######################
      if ($on=~/bsd|dragon|true/i) {    # bsd options
            $t=`zpool list 2>&1`;
            if (!($t=~/\"not found\"/s)) { $haszfs="yes"; } # missing "zpool" or zfs installed

            $on="bsd";
            if (($t[0]=~/free/i) && (!($t[1]=~/bsd/i))) {
              $t[1]="$t[0] $t[1]";
            }
            $on.=",$t[1],$t[2]";                       # os kernel, cpu

            $fn="bsd";
            $rh="BSD:$hn";

            $os="$uname";
            $nc="$spath/tools/nc/bsd/nc";

            #sambaconfig
            $sc=&get_smb_conf;
      }

      ########################
      if ($on=~/Darwin/i) {        # OSX

           $t=`zpool list 2>&1`;
           if (!($t=~/\"not found\"/s)) { $haszfs="yes"; } # missing "zpool" or zfs installed

            $rh="OSX:$hn";
            $on.=",OSX,$t[2],$t[3]";        # kernel, cpu   Darwin xxx

            $os="macos";
            $os.=`sw_vers -productVersion`;              # macOS 12.6

            $nc="$spath/tools/nc/osx/nc";
       }

       ########################
       if ($on=~/Linux/i) {         # Linux options

          $t=`zpool list 2>&1`;
           if (!($t=~/\"not found\"/s)) { $haszfs="yes"; } # missing "zpool" or zfs installed

           # Proxmox
           if ($t[1]=~/pve/i) {
                $t[1]=~s/pve/pve Proxmox/;
                $rh="PVE:$hn";
           }

           $on.=",$t[1],$t[2]";         # kernel, cpu
           $os=$uname;                  # os, kernel, cpu

           if ($uname=~/86/) {
              $nc="$spath/tools/nc/lin/nc";
           }

           #sambaconfig
           $sc=&get_smb_conf;
       }

       ########################
       if ($on=~/Solaris/i) {    # Solaris and forks
              # uname -mrs ->  always  SunOS 5.11 i86pc

              $haszfs="yes";
              my $buf="131072";

              if (-f "/etc/release") {
                 $t=`cat /etc/release`;
                 $t=~s/\n.*//s;                    # first line only with OS
                 $t=~s/^ +//;                     # no leading spaces
                 $t=~s/ *\(.*//;                  # no (opt text) ex in OI (powered..

                 #Solaris                         Oracle Solaris 11.4 X 86
                 if ($t=~/Solaris/i) {
                   $on.=",$t";
                   $os=$t;
                   $rh="SOL:$hn";
                 }

                 # OmniOS                          OmniOS v11 r151046z
                 if ($t=~/OmniOS/i) {
                   $on.=",$t";
                   $os=$t;
                   $rh="OMN:$hn";
                 }

                 #SmartOS                         SmartOS 20240404T141941Z x86_64
                 if ($t=~/SmartOS/i) {
                   $on.=",$t";
                   $os=$t;
                   $rh="SMA:$hn";
                 }

                 #OI                             OpenIndiana Hipster 2023.10 (powered..
                 if ($t=~/OpenIndiana/i) {
                   $on.=",$t";
                   $os=$t;
                   $rh="OI:$hn";
                 }

                 if ($os eq "")          {        # other Illumos
                    $on.=",$t";
                    $os="$t";
                    $rh="ILL:$hn";
                 }

              } else {      # no /etc/release

                  $on.=",$uname";
                  $os=$uname;
                  $rh="OSO:$hn";                 # other Solaris ??
              }

              if ($uname=~/86/) {
                  $nc="$spath/tools/nc/sol/nc -b $buf";
              }

              $t=`svcs | grep smb/server`;      # kernelbased smb
              if ($t=~/online/s)  {
                 $sc="smb/server";
              } else {
                 #sambaconfig
                 $sc=&get_smb_conf;
              }

       } # /sol

  }  #/else non windows


  $os=lc($os);

  $rh=~s/\n//gs;
  $os=~s/\n//gs;
  $rh=~s/\n//gs;

  # callback
  if ($comm=~/callback/i) { $on.=",callback"; }

  $fn=$on;   # fn and  on the same

  print "\nzfs $haszfs\n";                  # zfs installed
  print "dn: $dn\n";
  print "on: $on\n";                        #  lx,bsd,callback   family name
  print "fn: $fn\n";                        #  lx,bsd
  print "hn: $hn\n";                         # hostname
  print "os: $os\n";
  print "rh: $rh\n";
  print "nc: $nc\n";
  print "sc: $sc\n";                         # samba conf
  print "auth: ".substr($pw,0,4)."..\n\n";
  print "timeout: $timeout\n";
  print "max_connections: $max_connections\n";
  print "timeout: $timeout\n";
  print "allowed_ip: $allowed_ip\n";
  print "upsrv: $upsrv\n";
  print "path: $path\n";

  #perl INC
  print "\n\Perl \@INC\n";
  foreach my $dir (@INC) {
      print "$dir, ";
  }

  # ok
  print "\nzpool list\n".`zpool list`;
  print "\n\nwaiting for client requests..\n";

  return();

}


sub my_add {           # add ("path/filename","data",);  max 50 lines
##############

    my $file = $_[0];
    my $data = $_[1];
    my @old;

    ##########################
    my $afile = "$spath/tmp/activity.log";        # updated on file write
    my $lock="$spath/tmp/log.lock";
    open (X,"<$lock");
      flock(X, LOCK_EX);
          #
          open (F,"<$file");                      # old file
                flock(F, LOCK_EX);
                @old=<F>;
          close (F);

        my @new;
        for (my $i = 0; $i < @old; $i++) {
             #
             # Check if the maximum limit (100 elements) is reached
             if ($i >= 100) {
                 # Stop iterating if the limit is reached
                 last;
             }

             # Copy the current element from @old to @new_array
             push(@new, $old[$i]);
         }


         my $t=join("",@new);
         open (PF, ">$file");
            flock(PF, LOCK_EX);
            print PF "$t\n$data";                 # programm must add \n if needed to allow added strings
         close (PF);

         $t=time();
         open (PF, ">$afile");
            flock(PF, LOCK_EX);
            print PF "$t";                        # last  access
         close (PF);

    close (X);                                    # close lock


    #########################

    return();
}

sub my_append {           #   write ("path/filename","data",);
##############
    my $file = $_[0];
    my $data = $_[1];
    my $afile = "$spath/tmp/activity.log";         # updated on file write

    open(my $fh,">>$file");
      flock($fh, LOCK_EX);
        print $fh "$data\n";
      flock($fh, LOCK_UN);
    close($fh);

    $t=time();
    open(my $fh,">$afile");
      flock($fh, LOCK_EX);
        print $fh "$t";
      flock($fh, LOCK_UN);
    close($fh);

    return();
}

sub my_write {           #   write ("path/filename","data",);
##############
    my $file = $_[0];
    my $data = $_[1];
    my $afile = "$spath/tmp/activity.log";         # updated on file write

    open(my $fh,">$file");
      flock($fh, LOCK_EX);
        print $fh "$data\n";
      flock($fh, LOCK_UN);
    close($fh);

    my $t=time();
    open(my $fh,">$afile");
      flock($fh, LOCK_EX);
        print $fh "$t";
      flock($fh, LOCK_UN);
    close($fh);

    return();
}

sub my_read {           # read   read("path/filename");    opt return array
##############
    my $file = $_[0];
    my $arr  = $_[1];

    my $data="";
    my @data=();

    if (-f $file) {
        #
        open(my $fh, "<$file");
          flock($fh, LOCK_EX);
             @data = <$fh>;
          flock($fh, LOCK_UN);
        close($fh);


    } else {
         $data="$file not found";
    }

    if ($arr=~/a/) {  return(@data); }

    $data=join("",@data);
    return($data);
}

############   fileage tempfile in s
sub my_age {
###########
    # any write or add updates tmpfolder/activity.log with a date() stamp.
    # to sync with monitor.pl
    # this sub reads activity.log and returns age in s

    my $file = "$spath/tmp/activity.log";
    if (!(-f $file)) { return "-1"; }     # file not found

    my $r=&my_read($file);  # last access in s
    $t=time();
    my $old=$t-$r;  # current - last in s

    return($old);
}


###########   # send data via curl
sub my_updata {                          #
###########    # par ip, filename, auth, data      (-ks -> no cert check, silent)
   # curl -ks -X POST -F 'upfilefilename' -F 'data=abc' 'https://example.com/upload.cgi'

    my $upsrv=$_[0];
    my $filename=$_[1];         # filename without path
    my $auth=$_[2];
    my $text_data=$_[3];

    #$text_data= uri_escape($text_data);        # only for html uploads via Tiny etc

     print "\n".__LINE__." upload filename=$filename to $upsrv\n";
     print "curl --connect-timeout 5 -k -X POST -F \"upfile=$filename\" -F \"auth=$auth\" -F \"data=$text_data\" https://$upsrv/cgi-bin/cs/cs_connect.pl\n\n";

     my $r=`curl --connect-timeout 5 -k -X POST -F "upfile=$filename" -F "auth=$auth" -F "data=$text_data" https://$upsrv/cgi-bin/cs/cs_connect.pl`;
     print __LINE__." uploaded data to $filename on https://$upsrv/cgi-bin/cs/cs_connect.pl\n$r\n";
     if ($r=~/ok:/) { return ("$filename"); } else { return ("upload error: $r"); }

}

###########
###########   # send file via curl
sub my_upload {                          #
###########    # par ip, filename, auth, data      (-ks -> no cert check, silent)
   # curl -ks -X POST -F 'upfilefilename' -F 'data=@/path/to/your/file.txt' 'https://example.com/upload.cgi'

    my $up=$_[0];                    # ip
    my $filename=$_[1];              # filename without path
    my $auth=$_[2];
    my $ip=$_[3];

    my $k=" -k";                     # allow private cert
    if ($upsrv=~/[a-z]/) {           # from app setting, 'ip' or hostname
        if ($upsrv ne "ip") {        # self signed cert
          $up=$upsrv;
          $k="";
        }
    } #

    if ($up eq "") { $up=$ip; }
    #$text_data= uri_escape($text_data);        # only for html uploads via Tiny etc

     print "\n".__LINE__." upload filename=$filename to $up via\n";
     print "curl --connect-timeout 10$k -X POST -F \"upfile=$filename\" -F \"auth=$auth\" -F \"data=\@$cpath/$filename\" https://$up/cgi-bin/cs/cs_connect.pl\n\n";

     my $r=`curl --connect-timeout 10$k -X POST -F "upfile=$filename" -F "auth=$auth" -F "data=\@$cpath/$filename" https://$up/cgi-bin/cs/cs_connect.pl`;
     print __LINE__." uploaded $filename to https://$up/cgi-bin/cs/cs_connect.pl\n$r\n";

     if ($r=~/ok:/) { return ("$filename"); } else { return ("upload error:"); }

}

###########   # get file content via curl
sub my_download {                          #
###########    # par ip, filename, auth, data      (-ks -> no cert check, silent)
    # curl -ks -X POST -F 'upfilefilename' -F 'data=@/path/to/your/file.txt' 'https://example.com/upload.cgi'
    # cmd for callback  rcmd::upsrv::filename::memberip::auth:::nocache::

    my $up=$_[0];
    my $filename=$_[1];              # filename without path
    my $auth=$_[2];
    my $ip=$_[3];

    my $k=" -k";                     # allow private cert
    if ($upsrv=~/[a-z]/) {           # from app setting, 'ip' or hostname
        if ($upsrv ne "ip") {        # self signed cert
          $up=$upsrv;
          $k="";
        }
    } #

    if ($up eq "") { $up=$ip; }                 # local Apache

    #$text_data= uri_escape($text_data);        # only for html uploads via Tiny etc

    my $r=`curl --connect-timeout 10$k -X POST -F "getfile=$filename" -F "auth=$auth" https://$up/cgi-bin/cs/cs_connect.pl`;
    chomp($r);

    $t=substr($r,0.70);
    print "\n".__LINE__." download rcmdfile=$filename from $up via\n";
    print "curl --connect-timeout 10$k -X POST -F \"getfile=$filename\" -F \"auth=$auth\" https://$up/cgi-bin/cs/cs_connect.pl\nresult=$t\n";
    $r=~s/<n>/\n/gs;  # multiline transfer?


    if (($r eq  "") || ($r=~/error:/s))  {  return ("my_download error: or no data<hr>$r"); } else { return ($r); }

}

####################
sub c_filename {            # cache or command filename without ip and extension .data or .rcmd
###################
   my $cmd=$_[0];           # zpool list ..          #
   $cmd=~s/^ +| +$//g;      # unneded spaces
   $cmd=~s/ *line:=.*//;

   my $mem=$_[1];           #member 192...          #

  # get filename ip_cmd_hash        # max 64 B  ex: 127.0.0.1_zfs_get_all_623924ecd6b9239f336cc73d7691a219f.data

  my $h  = sha256_hex($cmd);        # sha of cmd
  $t=substr($cmd,0,30);             # first 30 char of cmd
  my $cfn="$mem\_$t\_$h";            # (ip)_cmd_hash
  $cfn=substr($cfn,0,55); # shorten to 55, real cmd is verified via file.last
  #&mess($cfn);
  $cfn=~s/ +/_/gs;
  $cfn=~s/[^a-zA-Z0-9\.\_\-]//gs;
  return($cfn);
}

###########   # send large data via post and Tiny
sub my_post {                          #
###########    # par ip, filename,data

    my $ip=$_[0];
    my $filename=$_[1];
    my $text_data=$_[2];         # plain text or data must be encoded.
    $text_data= uri_escape($text_data);

    my $len=length($text_data);

   # if ($filename=~/\.stat/) {
   #     $filename=~s/\.stat.*/.stat/;
   # }

    #debug
    #$filename="abc.tmp";    #   -> my_ip~abc.tmp

    # Data escaped
    # use URI::Escape;

    # $data_decoded = uri_unescape($data_encoded);

    # use HTTP::Tiny;  # 0.08    # in mainscript

    # Define the URL of the Apache script
    my $url = "http://$ip/cgi-bin/admin.pl";

    print "654 sub my_post, must upload $len Bytes data to $filename\n";

    my $boundary = "Boundary_".time();  # boundary string

    # HTTP-Request
    my $http = HTTP::Tiny->new;
    my $response = $http->request('POST', $url, {
        headers => {
            'Content-Type' => "multipart/form-data; boundary=$boundary",
        },
        content => "--$boundary\r\n" .
                    "Content-Disposition: form-data; name=\"file\"; filename=\"fn.txt\"\r\n" .
                    "Content-Type: text/plain\r\n\r\n" .
                    "$text_data\r\n" .

                    "--$boundary\r\n" .    # add formelement
                    "Content-Disposition: form-data; name=\"action\"\r\n\r\n" .
                    "rgroup_data_$filename\r\n" .  # filename
                    "--$boundary--\r\n",

    });


    my $response_body = $response->{content};
    #$response_body   = uri_unescape($response_body);                   # reverse escape

    # process answer
    if ($response->{success}) {
        print "683 data uploadad, $response_body\n";
        return ($filename);
    } else {
        print "686 upload error: ", $response->{status} || "other error", " (599 is network connect timeout)\n$response_body\n";
        print `ping $ip`."\n";
        return ("upload error: $response->{status}");
    }

}

#####################
sub client_ip_allowed    {
####################
my $client_ip=$_[0];

        print "\n".__LINE__." check if $client_ip is allowed: ";
        if ($client_ip eq "127.0.0.1") {
               print "always\n";
               return (1);
        }

        if ($allowed_ip ne "") {
           $t=$allowed_ip ;
           $t=~s/\./\\./g;                    # respect .
           $t=~s/ *, */|/g;                   # allow commalist ex 192.2.1.1,172.16, (this must be part of ip)
           if (!($client_ip=~/$t/)) {
             print "no\n";
             return (0);                      # not allowed
           }
        }
        print "yes\n";
        return (1);
}                                             # allowed

sub get_mydate {                # setzt datumsvariablen
###########
 my ($sec,$min,$hour,$mday,$mon,$year,$wday,$ydat,$isdst) = localtime(time);
 $mon++;
 if ($sec < 10) { $sec = "0$sec"; }
 if ($min < 10) { $min = "0$min"; }
 if ($hour < 10) { $hour = "0$hour"; }
 if ($mday < 10) { $mday = "0$mday"; }
 if ($mon < 10) { $mon = "0$mon"; }
 $year=1900+$year;
 return("$year\.$mon\.$mday\.$hour\.$min\.$sec");
}

#
1;
